@@ -1,14 +1,14 @@
-{-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE FlexibleContexts   #-}
-{-# LANGUAGE TemplateHaskell    #-}
-{-# LANGUAGE ViewPatterns       #-}
-{-# LANGUAGE DeriveGeneric        #-}
-{-# LANGUAGE DeriveAnyClass       #-}
+{-# LANGUAGE DeriveAnyClass        #-}
+{-# LANGUAGE DeriveDataTypeable    #-}
+{-# LANGUAGE DeriveGeneric         #-}
+{-# LANGUAGE FlexibleContexts      #-}
+{-# LANGUAGE FlexibleInstances     #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE TemplateHaskell       #-}
 -- |
 -- Copyright   : (c) 2011, 2012 Benedikt Schmidt & Simon Meier
 -- License     : GPL v3 (see LICENSE)
 --
--- Maintainer  : Simon Meier <iridcode@gmail.com>
 -- Portability : GHC only
 --
 -- Facts used to formulate and reason about protocol execution.
@@ -27,6 +27,8 @@ module Theory.Model.Fact (
   , isLinearFact
   , isPersistentFact
   , isProtoFact
+  , isInFact
+  , isKLogFact
 
   , factTagName
   , showFactTag
@@ -40,6 +42,7 @@ module Theory.Model.Fact (
   , getFactVariables
   , getFactTerms
   , getFactAnnotations
+  , getFactTag
   , isTrivialFact
   , isSolveFirstFact
   , isSolveLastFact
@@ -77,6 +80,7 @@ module Theory.Model.Fact (
   , protoFact
   , protoFactAnn
   , annotateFact
+  , applyMacroInFact
 
   -- * NFact
   , NFact
@@ -86,6 +90,7 @@ module Theory.Model.Fact (
   , LNFact
   , unifyLNFactEqs
   , unifiableLNFacts
+  , newVariables
 
   -- * Pretty-Printing
 
@@ -111,6 +116,7 @@ import qualified Data.Set               as S
 
 import           Term.Unification
 import           Term.Rewriting.Norm
+import           Term.Macro
 
 import           Text.PrettyPrint.Class
 
@@ -177,10 +183,10 @@ instance Sized t => Sized (Fact t) where
 
 instance HasFrees t => HasFrees (Fact t) where
     foldFrees  f = foldMap  (foldFrees f)
-    foldFreesOcc f c fa = foldFreesOcc f ((show $ factTag fa):c) (factTerms fa)
+    foldFreesOcc f c fa = foldFreesOcc f (show (factTag fa):c) (factTerms fa)
     mapFrees   f = traverse (mapFrees f)
 
-instance Apply t => Apply (Fact t) where
+instance Apply s t => Apply s (Fact t) where
     apply subst = fmap (apply subst)
 
 
@@ -238,12 +244,12 @@ isKDXorFact _                   = False
 
 -- | converts a KU-Fact into a KD-Fact with the same terms
 convertKUtoKD :: LNFact -> LNFact
-convertKUtoKD (Fact KUFact a m) = (Fact KDFact a m)
+convertKUtoKD (Fact KUFact a m) = Fact KDFact a m
 convertKUtoKD f                 = f
 
 -- | converts a KD-Fact into a KU-Fact with the same terms
 convertKDtoKU :: LNFact -> LNFact
-convertKDtoKU (Fact KDFact a m) = (Fact KUFact a m)
+convertKDtoKU (Fact KDFact a m) = Fact KUFact a m
 convertKDtoKU f                 = f
 
 -- | Mark a fact as malformed.
@@ -291,24 +297,38 @@ protoFactAnn multi name an ts = Fact (ProtoFact multi name (length ts)) an ts
 annotateFact :: S.Set FactAnnotation -> Fact t -> Fact t
 annotateFact ann' (Fact tag ann ts) = Fact tag (S.union ann' ann) ts
 
+-- | Apply macros in fact
+applyMacroInFact :: [Macro] -> LNFact -> LNFact
+applyMacroInFact mcs (Fact tag annot terms) = let mTerms = map (applyMacros mcs) terms in
+                                              Fact tag annot mTerms
+
 -- Queries on facts
 -------------------
 
 -- | True iff the fact is a non-special protocol fact.
 isProtoFact :: Fact t -> Bool
-isProtoFact (Fact (ProtoFact _ _ _) _ _) = True
+isProtoFact (Fact ProtoFact {} _ _) = True
 isProtoFact _                            = False
 
+-- | True iff the fact is an input fact.
+isInFact :: Fact t -> Bool
+isInFact (Fact InFact _ _) = True
+isInFact _                 = False
+
+-- | True iff the fact is the non-special Log-fact K() from the intruder rule isend
+isKLogFact :: Fact t -> Bool
+isKLogFact f = isProtoFact f && (factTagName (factTag f) == "K")
+
 -- | View a protocol fact.
 protoFactView :: LNFact -> Maybe [LNTerm]
 protoFactView fa = case fa of
-    Fact (ProtoFact _ _ _) _ m -> Just m
+    Fact ProtoFact {} _ m -> Just m
     _                          -> Nothing
 
 -- | View a protocol or in fact.
 protoOrInFactView :: LNFact -> Maybe [LNTerm]
 protoOrInFactView fa = case fa of
-    Fact (ProtoFact _ _ _) _  m  -> Just m
+    Fact ProtoFact {} _  m  -> Just m
     Fact InFact            _ [m] -> Just [m]
     Fact InFact            _  _  -> errMalformed "protoOrInFactView" fa
     _                            -> Nothing
@@ -316,7 +336,7 @@ protoOrInFactView fa = case fa of
 -- | View a protocol or out fact.
 protoOrOutFactView :: LNFact -> Maybe [LNTerm]
 protoOrOutFactView fa = case fa of
-    Fact (ProtoFact _ _ _) _  m  -> Just m
+    Fact ProtoFact {} _  m  -> Just m
     Fact OutFact           _ [m] -> Just [m]
     Fact OutFact           _  _  -> errMalformed "protoOrOutFactView" fa
     _                            -> Nothing
@@ -366,17 +386,20 @@ factMultiplicity = factTagMultiplicity . factTag
 getFactTerms :: Fact t -> [t]
 getFactTerms (Fact _ _ ts) = ts
 
+getFactTag :: Fact t -> FactTag
+getFactTag (Fact tag _ _) = tag
+
 -- | Get the set of fact annotations
 getFactAnnotations :: Fact t -> S.Set FactAnnotation
 getFactAnnotations (Fact _ ann _) = ann
 
 -- | Whether the fact has been marked as 'solve first' for the heuristic
 isSolveFirstFact :: Fact t -> Bool
-isSolveFirstFact (Fact tag ann _) = SolveFirst `S.member` ann || (isPrefixOf "F_" $ factTagName tag)
+isSolveFirstFact (Fact tag ann _) = SolveFirst `S.member` ann || isPrefixOf "F_" (factTagName tag)
 
 -- | Whether the fact has been marked as 'solve last' for the heuristic
 isSolveLastFact :: Fact t -> Bool
-isSolveLastFact (Fact tag ann _)  = SolveLast `S.member` ann  || (isPrefixOf "L_" $ factTagName tag)
+isSolveLastFact (Fact tag ann _)  = SolveLast `S.member` ann  || isPrefixOf "L_" (factTagName tag)
 
 -- | Whether the fact should not have its source solved
 isNoSourcesFact :: Fact t -> Bool
@@ -424,11 +447,11 @@ unifyLNFactEqs eqs
 
 -- | 'True' iff the two facts are unifiable.
 unifiableLNFacts :: LNFact -> LNFact -> WithMaude Bool
-unifiableLNFacts fa1 fa2 = (not . null) <$> unifyLNFactEqs [Equal fa1 fa2]
+unifiableLNFacts fa1 fa2 = not . null <$> unifyLNFactEqs [Equal fa1 fa2]
 
 -- | Normalize all terms in the fact
 normFact :: LNFact -> WithMaude LNFact
-normFact (Fact h an ts) = reader $ \hnd -> (Fact h an (map (\term -> runReader (norm' term) hnd) ts))
+normFact (Fact h an ts) = reader $ \hnd -> Fact h an (map (\term -> runReader (norm' term) hnd) ts)
 
 -- | @matchLFact t p@ is a complete set of AC matchers for the term fact @t@
 -- and the pattern fact @p@.
@@ -443,12 +466,12 @@ matchFact t p =
 -- | Get "left" variant of a diff fact
 getLeftFact :: LNFact -> LNFact
 getLeftFact (Fact tag an ts) =
-   (Fact tag an (map getLeftTerm ts))
+   Fact tag an (map getLeftTerm ts)
 
 -- | Get "left" variant of a diff fact
 getRightFact :: LNFact -> LNFact
 getRightFact (Fact tag an ts) =
-   (Fact tag an (map getRightTerm ts))
+   Fact tag an (map getRightTerm ts)
 
 -- | Get all variables inside a fact
 getFactVariables :: LNFact -> [LVar]
@@ -464,10 +487,17 @@ isTrivialFact (Fact _ _ ts) = case ts of
       combine :: Maybe [LVar] -> Maybe [LVar] -> Maybe [LVar]
       combine Nothing    _        = Nothing
       combine (Just _ )  Nothing  = Nothing
-      combine (Just l1) (Just l2) = if noDuplicates l1 l2 then (Just (l1++l2)) else Nothing
+      combine (Just l1) (Just l2) = if noDuplicates l1 l2 then Just (l1++l2) else Nothing
 
       noDuplicates l1 l2 = S.null (S.intersection (S.fromList l1) (S.fromList l2))
 
+newVariables :: [LNFact] -> [LNFact] -> [LNTerm]
+newVariables prems concs = map varTerm $ S.toList newvars
+  where
+    newvars = S.difference concvars premvars
+    premvars = S.fromList $ concatMap getFactVariables prems
+    concvars = S.fromList $ concatMap getFactVariables concs
+
 ------------------------------------------------------------------------------
 -- Pretty Printing
 ------------------------------------------------------------------------------
@@ -519,4 +549,4 @@ prettyNFact = prettyFact prettyNTerm
 
 -- | Pretty print a 'LFact'.
 prettyLNFact :: Document d => LNFact -> d
-prettyLNFact fa = prettyFact prettyNTerm fa
+prettyLNFact = prettyFact prettyNTerm
